//========================================================================
// This program is distributed under the GPL license.
// Copyright 2004 Juha Nieminen
//========================================================================

// STONESIZE should be odd for better line symmetry.
#define STONESIZE 19

#include "BWBitmap.hh"

#include <vector>
#include <string>
#include <iostream>
#include <fstream>
#include <cstdio>
#include <cstdlib>
#include <cctype>

namespace
{
    BWBitmap font;
}

void drawSquare(BWBitmap& bm, unsigned midx, unsigned midy)
{
    bm.setLineWidth(1.5);
    bm.line(midx-STONESIZE/4, midy-STONESIZE/4,
            midx+STONESIZE/4, midy-STONESIZE/4);
    bm.line(midx+STONESIZE/4, midy-STONESIZE/4,
            midx+STONESIZE/4, midy+STONESIZE/4);
    bm.line(midx+STONESIZE/4, midy+STONESIZE/4,
            midx-STONESIZE/4, midy+STONESIZE/4);
    bm.line(midx-STONESIZE/4, midy+STONESIZE/4,
            midx-STONESIZE/4, midy-STONESIZE/4);
    bm.setLineWidth(1);
}

void drawTriangle(BWBitmap& bm, unsigned midx, unsigned midy)
{
    double p1x = midx, p1y = midy-STONESIZE/3.0;
    double p2x = midx-.866*STONESIZE/3.0, p2y = midy+.5*STONESIZE/3.0;
    double p3x = midx+.866*STONESIZE/3.0;

    bm.setLineWidth(1.5);
    bm.line(p1x, p1y, p2x, p2y);
    bm.line(p2x, p2y, p3x, p2y);
    bm.line(p3x, p2y, p1x, p1y);
    bm.setLineWidth(1);
}

bool createPNG(const std::string& filename,
               const std::vector<std::string> diagram,
               bool whiteIsEven)
{
    unsigned width = ~0U;
    for(unsigned i=0; i<diagram.size(); ++i)
        if(diagram[i].size() < width) width = diagram[i].size();

    BWBitmap bm(width*STONESIZE, diagram.size()*STONESIZE);
    bm.setFont(font, 19, 19);

    for(unsigned line = 0; line < diagram.size(); ++line)
    {
        const unsigned y1 = line*STONESIZE, y2 = (line+1)*STONESIZE-1;
        const unsigned midy = y1+STONESIZE/2;
        for(unsigned row = 0; row < width; ++row)
        {
            const unsigned x1 = row*STONESIZE, x2 = (row+1)*STONESIZE-1;
            const unsigned midx = x1+STONESIZE/2;
            const char c = diagram[line][row];

            if((c >= '0' and c <= '9') or (c >= 'a' and c <='z'))
            {
                unsigned ind = c=='0' ? 9 : c<='9' ? c-'1' : c-'a'+10;
                if((whiteIsEven and ind%2 == 1) or
                   (!whiteIsEven and ind%2 == 0))
                {
                    bm.circle(midx, midy, STONESIZE/2);
                    bm.character(ind, midx, midy);
                }
                else
                {
                    bm.filledCircle(midx, midy, STONESIZE/2);
                    bm.setDrawingColor(255);
                    bm.character(ind, midx, midy);
                    bm.setDrawingColor(0);
                }
            }
            else if(c >= 'A' and c <= 'N')
            {
                bm.character(c-'A'+36, midx, midy);
                if(row > 0) bm.line(x1, midy, x1+1, midy);
                if(row < width-1) bm.line(x2-1, midy, x2, midy);
                if(line > 0) bm.line(midx, y1, midx, y1+1);
                if(line < diagram.size()-1) bm.line(midx, y2-1, midx, y2);
            }
            else switch(c)
            {
              case 'Q':
                  drawTriangle(bm, midx, midy);
              case 'O':
                  bm.circle(midx, midy, STONESIZE/2);
                  break;

              case 'X':
                  bm.filledCircle(midx, midy, STONESIZE/2);
                  break;

              case '@':
                  bm.filledCircle(midx, midy, STONESIZE/2);
                  bm.setDrawingColor(255);
                  drawTriangle(bm, midx, midy);
                  bm.setDrawingColor(0);
                  break;

              case ':':
                  drawSquare(bm, midx, midy);
                  goto CROSS;

              case ';':
                  drawSquare(bm, midx, midy);
              case ',':
                  bm.filledCircle(midx, midy, 2.2);
            CROSS:
              case '.':
                  bm.line(x1, midy, x2, midy);
                  bm.line(midx, y1, midx, y2);
                  break;

              case '=':
                  drawSquare(bm, midx, midy);
              case '-':
                  bm.line(x1, midy, x2, midy);
                  if(line == 0) bm.line(midx, midy, midx, y2);
                  else bm.line(midx, y1, midx, midy);
                  break;

              case '!':
                  drawSquare(bm, midx, midy);
              case '|':
                  bm.line(midx, y1, midx, y2);
                  if(row == 0) bm.line(midx, midy, x2, midy);
                  else bm.line(x1, midy, midx, midy);
                  break;

              case '#':
                  drawSquare(bm, midx, midy);
              case '+':
                  if(line == 0) bm.line(midx, midy, midx, y2);
                  else bm.line(midx, y1, midx, midy);
                  if(row == 0) bm.line(midx, midy, x2, midy);
                  else bm.line(x1, midy, midx, midy);
                  break;
            }
        }
    }

    return bm.writePNG(filename);
}

void printWarning(unsigned lineNumber, const std::string& msg)
{
    std::cerr << "Warning (line " << lineNumber << "): " << msg << ".\n";
}

unsigned parseDiagrams(std::istream& iFile)
{
    std::string line, filename;
    std::vector<std::string> diagram;
    unsigned lineNumber = 0, diagramsCreated = 0;

    while(true)
    {
        std::getline(iFile, line);
        if(!iFile.good()) return diagramsCreated;
        ++lineNumber;

        if(line.compare(0, 16, "<!-- GoDiagram \"") == 0)
        {
            unsigned i = 16;
            while(i < line.size() and line[i] != '"') ++i;
            filename = line.substr(16, i-16);
            if(filename.empty())
            {
                printWarning(lineNumber, "Empty file name");
                continue;
            }

            bool whiteIsEven = true;
            for(unsigned i = line.size()-1; true; --i)
                if(!std::isspace(line[i]))
                {
                    if(line[i] == '*') whiteIsEven = false;
                    break;
                }

            diagram.clear();

            while(true)
            {
                std::getline(iFile, line);
                if(!iFile.good())
                {
                    printWarning(lineNumber, "Premature end of string");
                    return diagramsCreated;
                }
                ++lineNumber;

                if(line.compare(0, 3, "-->") == 0)
                {
                    if(diagram.empty())
                    {
                        printWarning(lineNumber, "Empty diagram");
                    }
                    else
                    {
                        if(createPNG(filename, diagram, whiteIsEven))
                        {
                            std::cerr << "Created PNG file '" << filename
                                      << "'\n";
                            ++diagramsCreated;
                        }
                        else
                        {
                            printWarning(lineNumber,
                                         std::string("Couldn't create '")+
                                         filename+"'");
                        }
                    }
                    break;
                }
                else
                {
                    diagram.push_back(std::string());
                    diagram.back().reserve(line.size());
                    for(unsigned i=0; i<line.size(); ++i)
                        if(!std::isspace(line[i]))
                            diagram.back() += line[i];
                    if(diagram.back().empty())
                        diagram.pop_back();
                }
            }
        }
    }
}

int main(int argc, char* argv[])
{
    if(argc != 2)
    {
        std::cerr << "Usage: " << argv[0] << " <html file>\n";
        return 1;
    }

    std::string fontPath(std::getenv("HOME"));
    fontPath += "/.html2godiagramsfont.png";
    if(!font.readPNG(fontPath))
    {
        std::cerr <<
            "Please make $HOME/.html2godiagramsfont.png a symbolic link to\n"
            "the html2godiagrams package font.png file.\n"
            "(For example: ln -s /full/path/to/font.png "
            "~/.html2godiagramsfont.png)\n";
        return 1;
    }

    std::ifstream iFile(argv[1]);
    if(!iFile.good())
    {
        std::cerr << "Can't open ";
        std::perror(argv[1]);
        return 1;
    }

    unsigned diagramsCreated = parseDiagrams(iFile);
    if(diagramsCreated == 0)
    {
        std::cerr << "Warning: File contained no diagrams.\n";
    }
    else
    {
        std::cerr << "Created " << diagramsCreated << " diagram"
                  << (diagramsCreated>1 ? "s" : "") << ".\n";
    }

    return 0;
}
